from __future__ import annotations

import shutil
import typing
from pathlib import Path
from typing import Any

from digitalhub.datastores.api import get_datastore
from digitalhub.entities.dataitem._base.entity import Dataitem
from digitalhub.utils.uri_utils import check_local_path

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.dataitem.table.spec import DataitemSpecTable
    from digitalhub.entities.dataitem.table.status import DataitemStatusTable


class DataitemTable(Dataitem):
    """
    DataitemTable class.
    """

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: DataitemSpecTable,
        status: DataitemStatusTable,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        self.spec: DataitemSpecTable
        self.status: DataitemStatusTable

    def as_df(
        self,
        file_format: str | None = None,
        engine: str | None = None,
        clean_tmp_path: bool = True,
        **kwargs,
    ) -> Any:
        """
        Read dataitem file (csv or parquet) as a DataFrame from spec.path.
        If the dataitem is not local, it will be downloaded to a temporary
        folder named tmp_dir in the project context folder.
        If clean_tmp_path is True, the temporary folder will be deleted after the
        method is executed.
        It's possible to pass additional arguments to the this function. These
        keyword arguments will be passed to the DataFrame reader function such as
        pandas's read_csv or read_parquet.

        Parameters
        ----------
        file_format : str
            Format of the file. (Supported csv and parquet).
        engine : str
            Dataframe framework, by default pandas.
        clean_tmp_path : bool
            If True, the temporary folder will be deleted.
        **kwargs : dict
            Keyword arguments passed to the read_df function.

        Returns
        -------
        Any
            DataFrame.
        """
        if engine is None:
            engine = "pandas"
        try:
            if check_local_path(self.spec.path):
                tmp_dir = None
                data_path = self.spec.path
            else:
                tmp_dir = self._context().root / "tmp_data"
                tmp_dir.mkdir(parents=True, exist_ok=True)
                data_path = self.download(destination=str(tmp_dir), overwrite=True)

            if Path(data_path).is_dir():
                files = [str(i) for i in Path(data_path).rglob("*") if i.is_file()]
                checker = files[0]
            else:
                checker = data_path

            extension = self._get_extension(checker, file_format)
            datastore = get_datastore("")

            return datastore.read_df(data_path, extension, engine, **kwargs)

        except Exception as e:
            raise e

        finally:
            # Delete tmp folder
            self._clean_tmp_path(tmp_dir, clean_tmp_path)

    def write_df(
        self,
        df: Any,
        extension: str | None = None,
        **kwargs,
    ) -> str:
        """
        Write DataFrame as parquet/csv/table into dataitem spec.path.
        keyword arguments will be passed to the DataFrame reader function such as
        pandas's to_csv or to_parquet.

        Parameters
        ----------
        df : Any
            DataFrame to write.
        extension : str
            Extension of the file.
        **kwargs : dict
            Keyword arguments passed to the write_df function.

        Returns
        -------
        str
            Path to the written dataframe.
        """
        datastore = get_datastore(self.spec.path)
        return datastore.write_df(df, self.spec.path, extension=extension, **kwargs)

    @staticmethod
    def _clean_tmp_path(pth: Path | None, clean: bool) -> None:
        """
        Clean temporary path.

        Parameters
        ----------
        pth : Path | None
            Path to clean.
        clean : bool
            If True, the path will be cleaned.

        Returns
        -------
        None
        """
        if pth is not None and clean:
            shutil.rmtree(pth)



from __future__ import annotations

from pathlib import Path

from digitalhub.entities.utils.entity_types import EntityTypes
from digitalhub.utils.file_utils import get_file_mime_type
from digitalhub.utils.s3_utils import get_s3_bucket
from digitalhub.utils.uri_utils import check_local_path


def parse_entity_key(key: str) -> tuple[str]:
    """
    Parse the entity key. Returns project, entity type, kind, name and uuid.

    Parameters
    ----------
    key : str
        The entity key.

    Returns
    -------
    tuple[str]
        Project, entity type, kind, name and uuid.
    """
    try:
        # Remove "store://" from the key
        key = key.replace("store://", "")

        # Split the key into parts
        parts = key.split("/")

        # The project is the first part
        project = parts[0]

        # The entity type is the second part
        entity_type = parts[1]

        # The kind is the third part
        kind = parts[2]

        # Tasks and runs have no name and uuid
        if entity_type in (EntityTypes.TASK.value, EntityTypes.RUN.value):
            name = None
            uuid = parts[3]

        # The name and uuid are separated by a colon in the last part
        else:
            name, uuid = parts[3].split(":")

        return project, entity_type, kind, name, uuid
    except Exception as e:
        raise ValueError("Invalid key format.") from e


def eval_local_source(source: str | list[str]) -> None:
    """
    Evaluate if source is local.

    Parameters
    ----------
    source : str | list[str]
        Source(s).

    Returns
    -------
    None
    """
    if isinstance(source, list):
        if not source:
            raise ValueError("Empty list of sources.")
        source_is_local = all(check_local_path(s) for s in source)
        for s in source:
            if Path(s).is_dir():
                raise ValueError(f"Invalid source path: {s}. List of paths must be list of files, not directories.")
    else:
        source_is_local = check_local_path(source)

    if not source_is_local:
        raise ValueError("Invalid source path. Source must be a local path.")


def eval_zip_type(source: str | list[str]) -> bool:
    """
    Evaluate zip type.

    Parameters
    ----------
    source : str | list[str]
        Source(s).

    Returns
    -------
    bool
        True if path is zip.
    """
    if isinstance(source, list):
        if len(source) > 1:
            return False
        else:
            path = source[0]
    else:
        if Path(source).is_dir():
            return False
        path = source

    extension = path.endswith(".zip")
    mime_zip = get_file_mime_type(path) == "application/zip"
    return extension or mime_zip


def build_log_path_from_filename(
    project: str,
    entity_type: str,
    name: str,
    uuid: str,
    filename: str,
) -> str:
    """
    Build log path.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    name : str
        Object name.
    uuid : str
        Object UUID.
    filename : str
        Filename.

    Returns
    -------
    str
        Log path.
    """
    return f"s3://{get_s3_bucket()}/{project}/{entity_type}/{name}/{uuid}/{filename}"


def build_log_path_from_source(
    project: str,
    entity_type: str,
    name: str,
    uuid: str,
    source: str | list[str],
) -> str:
    """
    Build log path.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    name : str
        Object name.
    uuid : str
        Object UUID.
    source : str | list[str]
        Source(s).

    Returns
    -------
    str
        Log path.
    """
    is_zip = eval_zip_type(source)
    scheme = "zip+s3" if is_zip else "s3"
    path = f"{scheme}://{get_s3_bucket()}/{project}/{entity_type}/{name}/{uuid}"

    if isinstance(source, list) and len(source) >= 1:
        if len(source) > 1:
            path += "/"
        else:
            path += f"/{Path(source[0]).name}"
    elif Path(source).is_dir():
        path += "/"
    elif Path(source).is_file():
        path += f"/{Path(source).name}"

    return path


def get_entity_type_from_key(key: str) -> str:
    """
    Get entity type.

    Parameters
    ----------
    key : str
        The key of the entity.

    Returns
    -------
    str
        The entity type.
    """
    _, entity_type, _, _, _ = parse_entity_key(key)
    return entity_type



from __future__ import annotations

from digitalhub.entities.model._base.builder import ModelBuilder
from digitalhub.entities.model.huggingface.entity import ModelHuggingface
from digitalhub.entities.model.huggingface.spec import ModelSpecHuggingface, ModelValidatorHuggingface
from digitalhub.entities.model.huggingface.status import ModelStatusHuggingface


class ModelHuggingfaceBuilder(ModelBuilder):
    """
    ModelHuggingface builder.
    """

    ENTITY_CLASS = ModelHuggingface
    ENTITY_SPEC_CLASS = ModelSpecHuggingface
    ENTITY_SPEC_VALIDATOR = ModelValidatorHuggingface
    ENTITY_STATUS_CLASS = ModelStatusHuggingface
    ENTITY_KIND = "huggingface"



from __future__ import annotations

from pydantic import BaseModel, Field


class Signature(BaseModel):
    """
    MLFlow model signature.
    """

    inputs: str = None
    outputs: str = None
    params: str = None


class Dataset(BaseModel):
    """
    MLFlow model dataset.
    """

    name: str = None
    digest: str = None
    profile: str = None
    schema_: str = Field(default=None, alias="schema")
    source: str = None
    source_type: str = None



from __future__ import annotations

import typing

from digitalhub.context.builder import context_builder
from digitalhub.utils.exceptions import ContextError

if typing.TYPE_CHECKING:
    from digitalhub.context.context import Context
    from digitalhub.entities.project._base.entity import Project


def check_context(project: str) -> None:
    """
    Check if the given project is in the context.

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    bool
        True if the project is in the context, False otherwise.
    """
    if project not in context_builder._instances:
        raise ContextError


def set_context(project: Project) -> None:
    """
    Wrapper for ContextBuilder.build().

    Parameters
    ----------
    project : Project
        The project object used to set the current context.

    Returns
    -------
    None
    """
    context_builder.build(project)


def get_context(project: str) -> Context:
    """
    Wrapper for ContextBuilder.get().

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    Context
        The context for the given project name.
    """
    return context_builder.get(project)


def set_context_object(context: Context) -> None:
    """
    Wrapper for ContextBuilder.set().

    Parameters
    ----------
    context : Context
        The context to set.

    Returns
    -------
    None
    """
    context_builder.set(context)


def delete_context(project: str) -> None:
    """
    Wrapper for ContextBuilder.remove().

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    None
    """
    context_builder.remove(project)



from __future__ import annotations

import typing

from digitalhub.datastores.builder import builder

if typing.TYPE_CHECKING:
    from digitalhub.datastores._base.datastore import Datastore


def get_datastore(uri: str) -> Datastore:
    """
    Get a datastore instance by URI.

    Parameters
    ----------
    uri : str
        URI to parse.

    Returns
    -------
    Datastore
        The datastore instance.
    """
    return builder.get(uri)


def get_default_datastore() -> Datastore:
    """
    Get the default datastore instance.

    Returns
    -------
    Datastore
        The default datastore instance.
    """
    return builder.default()



from __future__ import annotations

import os

# Use env user as fallback in the API calls
try:
    FALLBACK_USER = os.getlogin()
except Exception:
    FALLBACK_USER = None

# File where to write DHCORE_ACCESS_TOKEN and DHCORE_REFRESH_TOKEN
# It's used because we inject the variables in jupyter env,
# but refresh token is only available once. Is it's used, we cannot
# overwrite it with coder, so we need to store the new one in a file,
# preserved for jupyter restart
ENV_FILE = ".dhcore"


# API levels that are supported
MAX_API_LEVEL = 20
MIN_API_LEVEL = 8



from __future__ import annotations

import time
import typing

from digitalhub.entities._base.api_utils import (
    list_entity_api_base,
    list_entity_api_ctx,
    logs_api,
    read_entity_api_ctx,
    resume_api,
    stop_api,
)
from digitalhub.entities._base.unversioned.entity import UnversionedEntity
from digitalhub.entities.utils.entity_types import EntityTypes
from digitalhub.entities.utils.state import State
from digitalhub.factory.api import (
    build_runtime,
    build_spec,
    build_status,
    get_entity_type_from_kind,
    get_executable_kind,
)
from digitalhub.utils.exceptions import EntityError
from digitalhub.utils.logger import LOGGER

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.run._base.spec import RunSpec
    from digitalhub.entities.run._base.status import RunStatus
    from digitalhub.runtimes._base import Runtime


class Run(UnversionedEntity):
    """
    A class representing a run.
    """

    ENTITY_TYPE = EntityTypes.RUN.value

    def __init__(
        self,
        project: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: RunSpec,
        status: RunStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, uuid, kind, metadata, spec, status, user)

        self.spec: RunSpec
        self.status: RunStatus

    ##############################
    #  Run Methods
    ##############################

    def build(self) -> None:
        """
        Build run.

        Returns
        -------
        None
        """
        executable = self._get_executable()
        task = self._get_task()
        new_spec = self._get_runtime().build(executable, task, self.to_dict())
        self.spec = build_spec(
            self.kind,
            **new_spec,
        )
        self._set_state(State.BUILT.value)
        self.save()

    def run(self) -> Run:
        """
        Run run.

        Returns
        -------
        Run
            Run object.
        """
        self.refresh()
        if self.spec.local_execution:
            if not self._is_ready_to_run():
                raise EntityError("Run is not in a state to run.")
            self._set_state(State.RUNNING.value)
            self.save(update=True)

        self._setup_execution()

        try:
            status = self._get_runtime().run(self.to_dict())
        except Exception as e:
            self.refresh()
            if self.spec.local_execution:
                self._set_state(State.ERROR.value)
            self._set_message(str(e))
            self.save(update=True)
            raise e

        self.refresh()
        if not self.spec.local_execution:
            status.pop("state", None)
        new_status = {**self.status.to_dict(), **status}
        self._set_status(new_status)
        self.save(update=True)
        return self

    def wait(self, log_info: bool = True) -> Run:
        """
        Wait for run to finish.

        Parameters
        ----------
        log_info : bool
            If True, log information.

        Returns
        -------
        Run
            Run object.
        """
        start = time.time()
        while True:
            if log_info:
                LOGGER.info(f"Waiting for run {self.id} to finish...")
            self.refresh()
            time.sleep(5)
            if self.status.state in [
                State.STOPPED.value,
                State.ERROR.value,
                State.COMPLETED.value,
            ]:
                if log_info:
                    current = time.time() - start
                    LOGGER.info(f"Run {self.id} finished in {current:.2f} seconds.")
                return self

    def logs(self) -> dict:
        """
        Get object from backend.
        Returns empty dictionary if context is local.

        Returns
        -------
        dict
            Logs from backend.
        """
        if self._context().local:
            return {}
        return logs_api(self.project, self.ENTITY_TYPE, self.id)

    def stop(self) -> None:
        """
        Stop run.

        Returns
        -------
        None
        """
        if not self._context().local and not self.spec.local_execution:
            return stop_api(self.project, self.ENTITY_TYPE, self.id)

    def resume(self) -> None:
        """
        Resume run.

        Returns
        -------
        None
        """
        if not self._context().local and not self.spec.local_execution:
            return resume_api(self.project, self.ENTITY_TYPE, self.id)
        self.run()

    ##############################
    #  Helpers
    ##############################

    def _setup_execution(self) -> None:
        """
        Setup run execution. In base class, nothing to do.

        Returns
        -------
        None
        """

    def _is_ready_to_run(self) -> bool:
        """
        Check if run is in a state ready for running (BUILT or STOPPED).

        Returns
        -------
        bool
            True if run is in runnable state, False otherwise.
        """
        return (self.status.state == State.BUILT.value) or (self.status.state == State.STOPPED.value)

    def _set_status(self, status: dict) -> None:
        """
        Set run status.

        Parameters
        ----------
        status : dict
            Status to set.

        Returns
        -------
        None
        """
        self.status: RunStatus = build_status(self.kind, **status)

    def _set_state(self, state: str) -> None:
        """
        Update run state.

        Parameters
        ----------
        state : str
            State to set.

        Returns
        -------
        None
        """
        self.status.state = state

    def _set_message(self, message: str) -> None:
        """
        Update run message.

        Parameters
        ----------
        message : str
            Message to set.

        Returns
        -------
        None
        """
        self.status.message = message

    def _get_runtime(self) -> Runtime:
        """
        Build runtime to build run or execute it.

        Returns
        -------
        Runtime
            Runtime object.
        """
        return build_runtime(self.kind, self.project)

    def _get_executable(self) -> dict:
        """
        Get executable object from backend. Reimplemented to avoid
        circular imports.

        Returns
        -------
        dict
            Executable (function or workflow) from backend.
        """
        exec_kind = get_executable_kind(self.kind)
        entity_type = get_entity_type_from_kind(exec_kind)
        splitted = self.spec.task.split("/")
        exec_name = splitted[-1].split(":")[0]
        exec_id = splitted[-1].split(":")[1]
        return read_entity_api_ctx(
            exec_name,
            entity_type=entity_type,
            project=self.project,
            entity_id=exec_id,
        )

    def _get_task(self) -> dict:
        """
        Get object from backend. Reimplemented to avoid
        circular imports.

        Returns
        -------
        dict
            Task from backend.
        """
        executable_kind = get_executable_kind(self.kind)
        exec_string = f"{executable_kind}://{self.spec.task.split('://')[1]}"

        # Local backend
        if self._context().local:
            tasks = list_entity_api_base(self._context().client, EntityTypes.TASK.value)
            for i in tasks:
                if i.get("spec").get("function") == exec_string:
                    return i
            raise EntityError("Task not found.")

        # Remote backend
        task_kind = self.spec.task.split("://")[0]
        params = {"function": exec_string, "kind": task_kind}
        return list_entity_api_ctx(self.project, EntityTypes.TASK.value, params=params)[0]



from __future__ import annotations

import typing

from digitalhub.client.api import build_client, get_client
from digitalhub.context.api import check_context
from digitalhub.entities._base.api_utils import (
    delete_entity_api_ctx,
    list_entity_api_ctx,
    read_entity_api_base,
    read_entity_api_ctx,
    read_entity_api_ctx_versions,
)
from digitalhub.entities.utils.entity_types import EntityTypes
from digitalhub.entities.utils.utils import parse_entity_key
from digitalhub.factory.api import build_entity_from_dict, build_entity_from_params
from digitalhub.utils.exceptions import ContextError, EntityAlreadyExistsError, EntityError, EntityNotExistsError
from digitalhub.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.context.entity import ContextEntity
    from digitalhub.entities._base.executable.entity import ExecutableEntity
    from digitalhub.entities._base.material.entity import MaterialEntity
    from digitalhub.entities._base.unversioned.entity import UnversionedEntity
    from digitalhub.entities._base.versioned.entity import VersionedEntity


def _check_context(project: str) -> None:
    """
    Check if the given project is in the context.
    Otherwise try to get the project from remote.

    Parameters
    ----------
    project : str
        Project name.

    Returns
    -------
    None
    """
    try:
        check_context(project)
    except ContextError:
        try:
            build_client()
            client = get_client()
            obj = read_entity_api_base(client, EntityTypes.PROJECT.value, project)
            build_entity_from_dict(obj)
        except EntityNotExistsError:
            raise ContextError(f"Project '{project}' not found.")


def _check_project_from_identifier(identifier: str, project: str | None = None) -> None:
    """
    Check if the given project is in the context.
    Otherwise try to get the project from remote.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.

    Returns
    -------
    None
    """
    if not identifier.startswith("store://"):
        if project is None:
            raise EntityError("Specify project if you do not specify entity key.")

    else:
        project, _, _, _, _ = parse_entity_key(identifier)

    _check_context(project)


def new_context_entity(**kwargs) -> ContextEntity:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    kind : str
        Kind the object.
    uuid : str
        ID of the object.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object spec must be embedded in project spec.

    Returns
    -------
    ContextEntity
        Object instance.
    """
    _check_context(kwargs["project"])
    obj = build_entity_from_params(**kwargs)
    obj.save()
    return obj


def get_versioned_entity(
    identifier: str,
    entity_type: str | None = None,
    project: str | None = None,
    entity_id: str | None = None,
    **kwargs,
) -> VersionedEntity:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    entity_type : str
        Entity type.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    VersionedEntity
        Object instance.
    """
    _check_project_from_identifier(identifier, project)
    obj = read_entity_api_ctx(
        identifier,
        entity_type=entity_type,
        project=project,
        entity_id=entity_id,
        **kwargs,
    )
    return build_entity_from_dict(obj)


def get_unversioned_entity(
    identifier: str,
    entity_type: str | None = None,
    project: str | None = None,
    **kwargs,
) -> UnversionedEntity:
    """
    Get object from backend.

    Parameters
    ----------
    entity_type : str
        Entity type.
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    UnversionedEntity
        Object instance.
    """
    _check_project_from_identifier(identifier, project)

    obj = read_entity_api_ctx(
        identifier,
        entity_type=entity_type,
        project=project,
        **kwargs,
    )
    return build_entity_from_dict(obj)


def get_material_entity(
    identifier: str,
    entity_type: str | None = None,
    project: str | None = None,
    entity_id: str | None = None,
    **kwargs,
) -> MaterialEntity:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    entity_type : str
        Entity type.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    MaterialEntity
        Object instance.
    """
    obj: MaterialEntity = get_versioned_entity(
        identifier,
        entity_type=entity_type,
        project=project,
        entity_id=entity_id,
        **kwargs,
    )
    obj._get_files_info()
    return obj


def get_context_entity_versions(
    identifier: str,
    entity_type: str | None = None,
    project: str | None = None,
    **kwargs,
) -> list[ContextEntity]:
    """
    Get object versions from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    entity_type : str
        Entity type.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[ContextEntity]
        List of object instances.
    """
    _check_project_from_identifier(identifier, project)
    obj = read_entity_api_ctx_versions(
        identifier,
        entity_type=entity_type,
        project=project,
        **kwargs,
    )
    return [build_entity_from_dict(o) for o in obj]


def get_material_entity_versions(
    identifier: str,
    entity_type: str | None = None,
    project: str | None = None,
    **kwargs,
) -> list[MaterialEntity]:
    """
    Get object versions from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    entity_type : str
        Entity type.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[MaterialEntity]
        List of object instances.
    """
    _check_project_from_identifier(identifier, project)
    objs = read_entity_api_ctx_versions(
        identifier,
        entity_type=entity_type,
        project=project,
        **kwargs,
    )
    objects = []
    for o in objs:
        entity: MaterialEntity = build_entity_from_dict(o)
        entity._get_files_info()
        objects.append(entity)
    return objects


def list_context_entities(project: str, entity_type: str, **kwargs) -> list[ContextEntity]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[ContextEntity]
        List of object instances.
    """
    _check_context(project)
    objs = list_entity_api_ctx(
        project=project,
        entity_type=entity_type,
        **kwargs,
    )
    return [build_entity_from_dict(obj) for obj in objs]


def list_material_entities(
    project: str,
    entity_type: str,
    **kwargs,
) -> list[MaterialEntity]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[MaterialEntity]
        List of object instances.
    """
    _check_context(project)
    objs = list_entity_api_ctx(
        project=project,
        entity_type=entity_type,
        **kwargs,
    )
    objects = []
    for o in objs:
        entity: MaterialEntity = build_entity_from_dict(o)
        entity._get_files_info()
        objects.append(entity)
    return objects


def import_context_entity(file: str) -> ContextEntity:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    ContextEntity
        Object instance.
    """
    dict_obj: dict = read_yaml(file)

    _check_context(dict_obj["project"])

    obj = build_entity_from_dict(dict_obj)
    try:
        obj.save()
    except EntityAlreadyExistsError:
        pass
    return obj


def import_executable_entity(file: str) -> ExecutableEntity:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    ExecutableEntity
        Object instance.
    """
    dict_obj: dict | list[dict] = read_yaml(file)
    if isinstance(dict_obj, list):
        exec_dict = dict_obj[0]
        tsk_dicts = dict_obj[1:]
    else:
        exec_dict = dict_obj
        tsk_dicts = []

    _check_context(exec_dict["project"])

    obj: ExecutableEntity = build_entity_from_dict(exec_dict)

    obj.import_tasks(tsk_dicts)

    try:
        obj.save()
    except EntityAlreadyExistsError:
        pass
    return obj


def delete_entity(
    identifier: str,
    entity_type: str | None = None,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
    cascade: bool = True,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    entity_type : str
        Entity type.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity. If True, use entity name instead of entity key as identifier.
    cascade : bool
        Cascade delete.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.
    """
    _check_project_from_identifier(identifier, project)
    return delete_entity_api_ctx(
        identifier=identifier,
        entity_type=entity_type,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
        cascade=cascade,
        **kwargs,
    )



from __future__ import annotations

import typing
from io import BytesIO
from typing import Any

from digitalhub.datastores._base.datastore import Datastore
from digitalhub.readers.api import get_reader_by_object

if typing.TYPE_CHECKING:
    from digitalhub.stores.s3.store import S3Store


class S3Datastore(Datastore):
    """
    S3 Datastore class.
    """

    def __init__(self, store: S3Store, **kwargs) -> None:
        super().__init__(store, **kwargs)
        self.store: S3Store

    def write_df(self, df: Any, dst: str, extension: str | None = None, **kwargs) -> str:
        """
        Write a dataframe to S3 based storage. Kwargs are passed to df.to_parquet().

        Parameters
        ----------
        df : Any
            The dataframe.
        dst : str
            The destination path on S3 based storage.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        str
            The S3 path where the dataframe was saved.
        """
        fileobj = BytesIO()
        reader = get_reader_by_object(df)
        reader.write_df(df, fileobj, extension=extension, **kwargs)

        key = self.store._get_key(dst)
        return self.store.upload_fileobject(fileobj, key)



from __future__ import annotations

import typing

from digitalhub.stores.builder import store_builder

if typing.TYPE_CHECKING:
    from digitalhub.stores._base.store import Store, StoreParameters


def set_store(store_cfg: StoreParameters) -> None:
    """
    Set a new store instance with the given configuration.

    Parameters
    ----------
    store_cfg : StoreParameters
        Store configuration.

    Returns
    -------
    None
    """
    store_builder.build(store_cfg)


def get_store(uri: str) -> Store:
    """
    Get store instance by uri.

    Parameters
    ---------
    uri : str
        URI to parse.

    Returns
    -------
    Store
        Store instance.
    """
    return store_builder.get(uri)


def get_default_store() -> Store:
    """
    Get the default store instance. The default store is the one that
    can persist artifacts and dataitems.

    Returns
    -------
    Store
        Default store instance.
    """
    return store_builder.default()



from __future__ import annotations

import typing

from digitalhub.entities._base.context.entity import ContextEntity
from digitalhub.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities._base.entity.spec import Spec
    from digitalhub.entities._base.entity.status import Status


class UnversionedEntity(ContextEntity):
    def __init__(
        self,
        project: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: Spec,
        status: Status,
        user: str | None = None,
    ) -> None:
        super().__init__(project, kind, metadata, spec, status, user)
        self.id = uuid
        self.key = f"store://{project}/{self.ENTITY_TYPE}/{kind}/{uuid}"
        self._obj_attr.extend(["id"])

    def export(self, filename: str | None = None) -> str:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        str
            Exported file.
        """
        obj = self.to_dict()
        if filename is None:
            filename = f"{self.ENTITY_TYPE}-{self.id}.yml"
        pth = self._context().root / filename
        write_yaml(pth, obj)
        return str(pth)



from __future__ import annotations

import typing

from digitalhub.datastores.local.datastore import LocalDatastore
from digitalhub.datastores.remote.datastore import RemoteDatastore
from digitalhub.datastores.s3.datastore import S3Datastore
from digitalhub.datastores.sql.datastore import SqlDatastore
from digitalhub.stores.api import get_default_store, get_store
from digitalhub.utils.uri_utils import map_uri_scheme

if typing.TYPE_CHECKING:
    from digitalhub.datastores._base.datastore import Datastore
    from digitalhub.stores._base.store import Store


REGISTRY_DATASTORES = {
    "local": LocalDatastore,
    "remote": RemoteDatastore,
    "s3": S3Datastore,
    "sql": SqlDatastore,
}


class DatastoreBuilder:
    """
    Datastore builder class.
    """

    def __init__(self) -> None:
        self._instances: dict[str, Datastore] = {}
        self._default: Datastore | None = None
        self._def_scheme = "s3"

    def build(self, uri: str) -> None:
        """
        Build a datastore instance and register it.
        It overrides any existing instance.

        Parameters
        ----------
        uri : str
            URI to parse.

        Returns
        -------
        None
        """
        store = get_store(uri)
        self._instances[store.type] = self.build_datastore(store)

    def get(self, uri: str) -> Datastore:
        """
        Get a store instance by URI.

        Parameters
        ----------
        uri : str
            URI to parse.

        Returns
        -------
        Datastore
            The datastore instance.
        """
        scheme = map_uri_scheme(uri)
        if scheme not in self._instances:
            self.build(uri)
        return self._instances[scheme]

    def default(self) -> Datastore:
        """
        Get the default store instance.

        Returns
        -------
        Datastore
            The default datastore instance.
        """
        if self._default is None:
            store = get_default_store()
            self._default = self.build_datastore(store)
        return self._default

    def build_datastore(self, store: Store) -> Datastore:
        """
        Build a datastore instance.

        Parameters
        ----------
        store : Store
            Store instance.

        Returns
        -------
        Datastore
            The datastore instance.

        Raises
        ------
        NotImplementedError
            If the store type is not implemented.
        """
        try:
            return REGISTRY_DATASTORES[store.type](store)
        except KeyError as e:
            raise NotImplementedError from e


builder = DatastoreBuilder()



from __future__ import annotations

import typing
from typing import Any

from digitalhub.datastores._base.datastore import Datastore

if typing.TYPE_CHECKING:
    from digitalhub.stores.remote.store import RemoteStore


class RemoteDatastore(Datastore):
    """
    Remote Datastore class.
    """

    def __init__(self, store: RemoteStore, **kwargs) -> None:
        super().__init__(store, **kwargs)
        self.store: RemoteStore

    def write_df(self, df: Any, dst: str, extension: str | None = None, **kwargs) -> str:
        """
        Method to write a dataframe to a file. Note that this method is not implemented
        since the remote store is not meant to write dataframes.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        raise NotImplementedError("Remote store does not support write_df.")



import dotenv

import digitalhub
from digitalhub.entities.task._base.entity import Task

dotenv.load_dotenv()


def add_param(kwargs) -> dict:
    if kwargs["kind"] == "mlrun+job":
        kwargs["function"] = f1._get_function_string()

    if kwargs["kind"] == "dbt+transform":
        kwargs["function"] = f2._get_function_string()

    if kwargs["kind"] == "container+job":
        kwargs["function"] = f3._get_function_string()

    if kwargs["kind"] == "nefertem+infer":
        kwargs["function"] = f4._get_function_string()
        kwargs["framework"] = "test"

    return kwargs


names = ["test1", "test2", "test3", "test4"]
uuids = [
    "12a01efa-o44f-4991-b153-9a3c358b7bb0",
    "8e367f52-25bb-4df1-b9c9-a58045b377a0",
    "1678f9ab-a2e0-48ff-870a-2384o3fa1334",
    "adb746dd-4e81-4ff8-82de-4916624o17dc",
]
kind = ["mlrun+job", "dbt+transform", "container+job", "nefertem+infer"]

dicts = []
for i in range(len(names)):
    dicts.append({"name": names[i], "uuid": uuids[i], "kind": kind[i]})

digitalhub.delete_project("test")

p = digitalhub.get_or_create_project("test")

f1 = p.new_function(name="t1", kind="mlrun", source={"code": "test"})
f2 = p.new_function(name="t2", kind="dbt", source={"code": "test"})
f3 = p.new_function(name="t3", kind="container", image="test")
f4 = p.new_function(name="t4", kind="nefertem")


# Create and delete tasks
for i in dicts:
    i = add_param(i)
    d = digitalhub.new_task(p.name, **i)
    digitalhub.delete_task(p.name, entity_id=d.id)

# Create multiple tasks
for i in dicts:
    i = add_param(i)
    digitalhub.new_task(p.name, **i)

# List tasks
l_obj = digitalhub.list_tasks(p.name)
assert isinstance(l_obj, list)
assert len(l_obj) == 4
for i in l_obj:
    assert isinstance(i, dict)

for uuid in uuids:
    digitalhub.delete_task(p.name, entity_id=uuid)

# Get tasks test
for i in dicts:
    i = add_param(i)
    o1 = digitalhub.new_task(p.name, **i)
    assert isinstance(o1, Task)

    # Get by id
    o2 = digitalhub.get_task(p.name, entity_id=o1.id)
    assert isinstance(o2, Task)
    assert o1.id == o2.id

digitalhub.delete_project("test")



from __future__ import annotations

import typing
from pathlib import Path
from typing import Any

from digitalhub.datastores._base.datastore import Datastore
from digitalhub.readers.api import get_reader_by_object

if typing.TYPE_CHECKING:
    from digitalhub.stores.local.store import LocalStore


class LocalDatastore(Datastore):
    """
    Local Datastore class.
    """

    def __init__(self, store: LocalStore, **kwargs) -> None:
        super().__init__(store, **kwargs)
        self.store: LocalStore

    def write_df(self, df: Any, dst: str, extension: str | None = None, **kwargs) -> str:
        """
        Method to write a dataframe to a file. Kwargs are passed to df.to_parquet().
        If destination is not provided, the dataframe is written to the default
        store path with generated name.

        Parameters
        ----------
        df : Any
            The dataframe to write.
        dst : str
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        str
            Path of written dataframe.
        """
        self.store._check_local_dst(dst)
        self._validate_extension(Path(dst).suffix.removeprefix("."))

        # Write dataframe
        reader = get_reader_by_object(df)
        reader.write_df(df, dst, extension=extension, **kwargs)

        return dst



from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator
from digitalhub.entities.task._base.models import K8s


class TaskSpec(Spec):
    """TaskSpec specifications."""

    def __init__(self, function: str) -> None:
        self.function = function


class TaskSpecK8s(TaskSpec):
    """TaskSpecK8s specifications."""

    def __init__(
        self,
        function: str,
        node_selector: dict | None = None,
        volumes: list | None = None,
        resources: dict | None = None,
        affinity: dict | None = None,
        tolerations: list | None = None,
        envs: list | None = None,
        secrets: list | None = None,
        profile: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(function)
        self.node_selector = node_selector
        self.volumes = volumes
        self.resources = resources
        self.affinity = affinity
        self.tolerations = tolerations
        self.envs = envs
        self.secrets = secrets
        self.profile = profile


class TaskValidator(SpecValidator):
    """
    TaskValidator validator.
    """

    function: str
    """Function string."""


class TaskValidatorK8s(TaskValidator, K8s):
    """
    TaskValidatorK8s validator.
    """



from __future__ import annotations

from io import BytesIO
from typing import Any

import numpy as np
import pandas as pd
from pandas.errors import ParserError

from digitalhub.readers._base.reader import DataframeReader
from digitalhub.utils.data_utils import build_data_preview, get_data_preview


class DataframeReaderPandas(DataframeReader):
    """
    Pandas reader class.
    """

    ##############################
    # Read methods
    ##############################

    def read_df(self, path: str | list[str], extension: str, **kwargs) -> pd.DataFrame:
        """
        Read DataFrame from path.

        Parameters
        ----------
        path : str | list[str]
            Path(s) to read DataFrame from.
        extension : str
            Extension of the file.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame.
        """
        if extension == "csv":
            method = pd.read_csv
        elif extension == "parquet":
            method = pd.read_parquet
        elif extension == "file":
            try:
                return self.read_df(path, "csv", **kwargs)
            except ParserError:
                raise ValueError(f"Unable to read from {path}.")

        if isinstance(path, list):
            dfs = [method(p, **kwargs) for p in path]
            return pd.concat(dfs)
        return method(path, **kwargs)

    ##############################
    # Write methods
    ##############################

    def write_df(self, df: pd.DataFrame, dst: str | BytesIO, extension: str | None = None, **kwargs) -> None:
        """
        Write DataFrame as parquet.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        dst : str | BytesIO
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None
        """
        if extension == "csv":
            return self.write_csv(df, dst, **kwargs)
        self.write_parquet(df, dst, **kwargs)

    @staticmethod
    def write_csv(df: pd.DataFrame, dst: str | BytesIO, **kwargs) -> None:
        """
        Write DataFrame as csv.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        dst : str | BytesIO
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None
        """
        df.to_csv(dst, index=False, **kwargs)

    @staticmethod
    def write_parquet(df: pd.DataFrame, dst: str | BytesIO, **kwargs) -> None:
        """
        Write DataFrame as parquet.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        dst : str | BytesIO
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None
        """
        df.to_parquet(dst, index=False, **kwargs)

    @staticmethod
    def write_table(df: pd.DataFrame, table: str, engine: Any, schema: str, **kwargs) -> None:
        """
        Write DataFrame as table.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        table : str
            The destination table.
        engine : Any
            The SQLAlchemy engine.
        schema : str
            The destination schema.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None
        """
        df.to_sql(table, engine, schema=schema, index=False, **kwargs)

    ##############################
    # Utils
    ##############################

    @staticmethod
    def get_schema(df: pd.DataFrame) -> Any:
        """
        Get schema.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe.

        Returns
        -------
        Any
            The schema.
        """
        schema = {"fields": []}

        for column_name, dtype in df.dtypes.items():
            field = {"name": str(column_name), "type": ""}

            if pd.api.types.is_integer_dtype(dtype):
                field["type"] = "integer"
            elif pd.api.types.is_float_dtype(dtype):
                field["type"] = "number"
            elif pd.api.types.is_bool_dtype(dtype):
                field["type"] = "boolean"
            elif pd.api.types.is_string_dtype(dtype):
                field["type"] = "string"
            elif pd.api.types.is_datetime64_any_dtype(dtype):
                field["type"] = "datetime"
            else:
                field["type"] = "any"

            schema["fields"].append(field)

        return schema

    @staticmethod
    def get_preview(df: pd.DataFrame) -> Any:
        """
        Get preview.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe.

        Returns
        -------
        Any
            The preview.
        """
        columns = [str(col) for col, _ in df.dtypes.items()]
        head = df.head(10)
        head = head.replace({np.nan: None})
        head = head.values.tolist()
        preview = get_data_preview(columns, head)
        len_df = len(df)
        return build_data_preview(preview, len_df)



from __future__ import annotations

from pydantic import BaseModel

from digitalhub.client.dhcore.env import FALLBACK_USER


class AuthConfig(BaseModel):
    """Client configuration model."""

    user: str = FALLBACK_USER
    """Username."""


class BasicAuth(AuthConfig):
    """Basic authentication model."""

    password: str
    """Basic authentication password."""


class ClientParams(AuthConfig):
    """Client id authentication model."""

    client_id: str = None
    """OAuth2 client id."""

    client_scecret: str = None
    """OAuth2 client secret."""


class OAuth2TokenAuth(ClientParams):
    """OAuth2 token authentication model."""

    access_token: str
    """OAuth2 token."""

    refresh_token: str = None
    """OAuth2 refresh token."""


class TokenExchangeAuth(ClientParams):
    """Token exchange authentication model."""

    exchange_token: str
    """Exchange token."""



from __future__ import annotations

import typing
from abc import ABCMeta, abstractmethod
from typing import Any

from digitalhub.readers.api import get_reader_by_engine

if typing.TYPE_CHECKING:
    from digitalhub.stores._base.store import Store


class Datastore(metaclass=ABCMeta):
    """
    Datastore abstract class.
    """

    def __init__(self, store: Store, **kwargs) -> None:
        self.store = store

    ##############################
    # IO methods
    ##############################

    @abstractmethod
    def write_df(self, df: Any, dst: str, extension: str | None = None, **kwargs) -> str:
        """
        Write DataFrame as parquet or csv.
        """

    def read_df(
        self,
        path: str | list[str],
        extension: str,
        engine: str | None = None,
        **kwargs,
    ) -> Any:
        """
        Read DataFrame from path.

        Parameters
        ----------
        path : str | list[str]
            Path(s) to read DataFrame from.
        extension : str
            Extension of the file.
        engine : str
            Dataframe engine (pandas, polars, etc.).
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Any
            DataFrame.
        """
        reader = get_reader_by_engine(engine)
        self._validate_extension(extension)
        return reader.read_df(path, extension, **kwargs)

    ##############################
    # Helper methods
    ##############################

    @staticmethod
    def _validate_extension(extension: str) -> None:
        """
        Validate extension.

        Parameters
        ----------
        extension : str
            Extension of the file.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If extension is not supported.
        """
        if extension not in ["csv", "parquet", "file"]:
            raise ValueError(f"Extension {extension} not supported.")



from __future__ import annotations

import typing

from digitalhub.entities._base.executable.entity import ExecutableEntity
from digitalhub.entities.utils.entity_types import EntityTypes
from digitalhub.factory.api import get_run_kind, get_task_kind_from_action
from digitalhub.utils.exceptions import BackendError

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.run._base.entity import Run
    from digitalhub.entities.workflow._base.spec import WorkflowSpec
    from digitalhub.entities.workflow._base.status import WorkflowStatus


class Workflow(ExecutableEntity):
    """
    A class representing a workflow.
    """

    ENTITY_TYPE = EntityTypes.WORKFLOW.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: WorkflowSpec,
        status: WorkflowStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)

        self.spec: WorkflowSpec
        self.status: WorkflowStatus

    ##############################
    #  Workflow Methods
    ##############################

    def run(self, action: str | None = None, **kwargs) -> Run:
        """
        Run workflow.

        Parameters
        ----------
        action : str
            Action to execute.
        **kwargs : dict
            Keyword arguments passed to Run builder.

        Returns
        -------
        Run
            Run instance.
        """
        if action is None:
            action = "pipeline"

        # Get task and run kind
        task_kind = get_task_kind_from_action(self.kind, action)
        run_kind = get_run_kind(self.kind)

        # Create or update new task
        task = self._get_or_create_task(task_kind)

        # Raise error if execution is not done by DHCore backend
        if self._context().local:
            raise BackendError("Cannot run workflow with local backend.")

        return task.run(run_kind, local_execution=False, **kwargs)



from __future__ import annotations

import json


def build_data_preview(preview: list[dict] | None = None, rows_count: int | None = None) -> dict:
    """
    Build data preview.

    Parameters
    ----------
    preview : list[dict] | None
        Preview.
    rows_count : int | None
        Row count.

    Returns
    -------
    dict
        Data preview.
    """
    dict_ = {}
    if preview is not None:
        dict_["cols"] = preview
    if rows_count is not None:
        dict_["rows"] = rows_count
    return dict_


def get_data_preview(columns: list, data: list[list], columnar: bool = False) -> list[dict]:
    """
    Prepare preview.

    Parameters
    ----------
    columns : list
        Columns names.
    data : list[list]
        Data to preview.
    columnar : bool
        If data are arranged in columns. If False, data are arranged in rows.

    Returns
    -------
    list[dict]
        Data preview.
    """
    # Reduce data to 10 rows
    if not columnar:
        if len(data) > 10:
            data = data[:10]
    else:
        data = [d[:10] for d in data]

    # Transpose data if needed
    if not columnar:
        data = list(map(list, list(zip(*data))))

    # Prepare the preview
    data_dict = prepare_preview(columns, data)

    # Filter memoryview values
    filtered_memview = filter_memoryview(data_dict)

    # Check the size of the preview data
    return check_preview_size(filtered_memview)


def prepare_preview(column_names: list, data: list[list]) -> list[dict]:
    """
    Get preview.

    Parameters
    ----------
    data : pd.DataFrame
        Data.

    Returns
    -------
    list[dict]
        Preview.
    """
    if len(column_names) != len(data):
        raise ValueError("Column names and data must have the same length")
    return [{"name": column, "value": values} for column, values in zip(column_names, data)]


def filter_memoryview(data: list[dict]) -> list[dict]:
    """
    Find memoryview values.

    Parameters
    ----------
    data : pd.DataFrame
        Data.

    Returns
    -------
    list[str]
        Column to filter out from preview.
    """
    key_to_filter = []
    for i in data:
        if any(isinstance(v, memoryview) for v in i["value"]):
            key_to_filter.append(i["name"])
    for i in key_to_filter:
        data = [d for d in data if d["name"] != i]
    return data


def check_preview_size(preview: list[dict]) -> list:
    """
    Check preview size. If it's too big, return empty list.

    Parameters
    ----------
    preview : list[dict]
        Preview.

    Returns
    -------
    list
        Preview.
    """
    if len(json.dumps(preview).encode("utf-8")) >= 64000:
        return []
    return preview



from __future__ import annotations

import typing
from typing import Any

from digitalhub.datastores._base.datastore import Datastore
from digitalhub.readers.api import get_reader_by_object

if typing.TYPE_CHECKING:
    from digitalhub.stores.sql.store import SqlStore


class SqlDatastore(Datastore):
    """
    Sql Datastore class.
    """

    def __init__(self, store: SqlStore, **kwargs) -> None:
        super().__init__(store, **kwargs)
        self.store: SqlStore

    def write_df(self, df: Any, dst: str, extension: str | None = None, **kwargs) -> str:
        """
        Write a dataframe to a database. Kwargs are passed to df.to_sql().

        Parameters
        ----------
        df : Any
            The dataframe to write.
        dst : str
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        str
            Path of written dataframe.
        """
        schema = self.store._get_schema(dst)
        table = self.store._get_table_name(dst)
        return self._upload_table(df, schema, table, **kwargs)

    def _upload_table(self, df: Any, schema: str, table: str, **kwargs) -> str:
        """
        Upload a table to SQL based storage.

        Parameters
        ----------
        df : DataFrame
            The dataframe.
        schema : str
            Destination schema.
        table : str
            Destination table.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        str
            The SQL URI where the dataframe was saved.
        """
        reader = get_reader_by_object(df)
        engine = self.store._check_factory()
        reader.write_table(df, table, engine, schema, **kwargs)
        engine.dispose()
        return f"sql://{engine.url.database}/{schema}/{table}"



from __future__ import annotations

import typing
from pathlib import Path

from digitalhub.entities._base.material.entity import MaterialEntity
from digitalhub.entities.utils.entity_types import EntityTypes
from digitalhub.utils.exceptions import EntityError
from digitalhub.utils.uri_utils import map_uri_scheme

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.dataitem._base.spec import DataitemSpec
    from digitalhub.entities.dataitem._base.status import DataitemStatus


class Dataitem(MaterialEntity):
    """
    A class representing a dataitem.
    """

    ENTITY_TYPE = EntityTypes.DATAITEM.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: DataitemSpec,
        status: DataitemStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)
        self.spec: DataitemSpec
        self.status: DataitemStatus

    ##############################
    #  Private helper methods
    ##############################

    @staticmethod
    def _get_extension(path: str, file_format: str | None = None) -> str:
        """
        Get extension of path.

        Parameters
        ----------
        path : str
            Path to get extension from.
        file_format : str
            File format.

        Returns
        -------
        str
            File extension.

        Raises
        ------
        EntityError
            If file format is not supported.
        """
        if file_format is not None:
            return file_format

        scheme = map_uri_scheme(path)
        if scheme == "sql":
            return "parquet"

        ext = Path(path).suffix[1:]
        if ext is not None:
            return ext
        raise EntityError("Unknown file format. Only csv and parquet are supported.")
